# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program is a sample from the 'Hack the Badge' workshop

Implementation Notes
--------------------
**Hardware:**
    * one of the Explorer Badge designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries

**Open Source Repository:**
    * GitLab: https://gitlab.com/bradanlane/exp_tutorial
    
'''

import time
import board


class expEEPROM:
    '''
        The CAT24C16W is a 2KB EEPROM with I2C interface.
            - I2C is a 2-wire serial interface which allows many devices to share the same wires.
            - Each I2C device, sharing the I2C wires, has its own 1-byte address.
            - CircuitPython provides a means to scan I2C for all attached devices.

        The EEPROM has a series of I2C addresses starting with `0x50` and is accessed using `board.I2C()`.
        The 2KB is broken into 8 256 byte pages.
        
        Reads and writes can be to specific addresses or - by setting 'position' they can be sequential.
    '''
    
    def __init__(self) -> None:
        self._max_size = 0x800  # 2KB
        self._page_count = 8
        self._current_position = 0  # only support sequential reads or writes
    
        self._i2c = board.I2C()
        self._base_address = 0x50    # each 'address' is 256 bytes; a total of 8 addresses are used
        # thee are created here to avoid repeated memory allocations
        self._data = bytearray(1)
        self._read_cmd = bytearray(1)
        self._write_cmd = bytearray(2)
        
    def __len__(self) -> int:
        return self._max_size

    def _read_eeprom(self, count: int, fixed=None) -> int:
        ''' internal method to read a byte or word '''
        if count not in (1, 2):
            raise ValueError("Read 1 or 2 bytes only")
        if fixed is not None:
            address = fixed
        else:
            address = self._current_position
        try:
            if self._i2c.try_lock():
                block = address // 256
                if block >= self._page_count:
                    raise ValueError("Block out of range")
                offset = address % 256
                self._read_cmd[0] = offset
                self._i2c.writeto(block + self._base_address, self._read_cmd)
                self._i2c.readfrom_into(self._base_address, self._data)
                val = self._data[0]
                if count == 2:
                    address += 1
                    block = address // 256
                    if block >= self._page_count:
                        raise ValueError("Block out of range")
                    offset = address % 256
                    self._read_cmd[0] = offset
                    self._i2c.writeto(block + self._base_address, self._read_cmd)
                    self._i2c.readfrom_into(self._base_address, self._data)
                    val |= self._data[0] << 8
                self._i2c.unlock()
            else:
                self._i2c.unlock()
                print ("Rats! The storage is locked and I can't find the key")
                val = -1
        except Exception as e:
            self._i2c.unlock()
            print ("EEEK! I can't read what was written. Something about ", e)
            val = -1
        if fixed is None:
            self._current_position += count
        #print (f"Read {count} bytes = {val}")   # DELETEME
        return val

    def __write_eeprom(self, count: int, val: int, fixed=None) -> int:
        ''' internal method to write a byte or word '''
        if count not in (1, 2):
            raise ValueError("Read 1 or 2 bytes only")

        if fixed is not None:
            address = fixed
        else:
            address = self._current_position
        try:
            if self._i2c.try_lock():
                # command and data must be sent together for proper timing
                block = address // 256
                if block >= self._page_count:
                    raise ValueError("Block out of range")
                offset = address % 256
                self._write_cmd[0] = offset
                self._write_cmd[1] = val & 0xFF
                self._i2c.writeto(block + self._base_address, self._write_cmd)
                time.sleep(0.006) # 5ms write cycle time plus a little extra
                if count == 2:
                    next_address = address + 1
                    block = next_address // 256
                    if block >= self._page_count:
                        raise ValueError("Block out of range")
                    offset = next_address % 256
                    self._write_cmd[0] = offset
                    self._write_cmd[1] = (val >> 8) & 0xFF
                    self._i2c.writeto(block + self._base_address, self._write_cmd)
                    time.sleep(0.006)
                self._i2c.unlock()
                result = self._read_eeprom(count, fixed=address)
            else:
                print ("What?! The storage is locked and I can't find the key")
                self._i2c.unlock()
                result = -1
        except Exception as e:
            print ("EEEK! I have forgotten how to write. Something about ", e)
            self._i2c.unlock()
            result = -1
        if fixed is None:
            self._current_position += count
        #print (f"Write {count} bytes = {val} == {result}") # DELETEME
        return val

    def readByte(self, addr=None) -> int:
        if addr is not None:
            return self._read_eeprom(1, fixed=addr)
        return self._read_eeprom(1)
        
    def readWord(self, addr=None) -> int:
        if addr is not None:
            return self._read_eeprom(2, fixed=addr)
        return self._read_eeprom(2)

    def writeByte(self, val: int, addr=None) -> int:
        if addr is not None:
            return self.__write_eeprom(1, val, fixed=addr)
        return self.__write_eeprom(1, val)
    
    def writeWord(self, val: int, addr=None) -> int:
        if addr is not None:
            return self.__write_eeprom(2, val, fixed=addr)
        return self.__write_eeprom(2, val)

    @property
    def position(self) -> int:
        return self._current_position
    @position.setter
    def position(self, value: int) -> None:
        self._current_position = value

MARKER_ADDRESS = 0x00
MARKER_VALUE = 0xA0
DATA_ADDRESS = 0x10

eeprom = expEEPROM()

MESSAGE = "Hello, World!"

if eeprom.readByte(addr=MARKER_ADDRESS) != MARKER_VALUE:
    eeprom.writeByte(MARKER_VALUE, addr=MARKER_ADDRESS)
    print("The EERPOM has not been used before")
    eeprom.writeByte(1, addr=DATA_ADDRESS)

    print("Writing a message to the EEPROM")
    eeprom.position = DATA_ADDRESS + 1
    eeprom.writeByte(len(MESSAGE))

    message_bytes = bytearray(MESSAGE, 'utf-8')
    for byte in message_bytes:
        eeprom.writeByte(byte)
else:
    val = eeprom.readByte(addr=DATA_ADDRESS)
    print(f"The EERPOM been used {val:d} time(s) before")
    val += 1
    eeprom.writeByte(val, addr=DATA_ADDRESS)
    print("Reading a message from the EEPROM")
    eeprom.position = DATA_ADDRESS + 1
    count = eeprom.readByte()
    message_bytes = bytearray()
    for i in range(count):
        message_bytes.append(eeprom.readByte())
    print("".join(map(chr, message_bytes)))
